#ifndef CPSASM_COMMON_H
#define CPSASM_COMMON_H

// max. nr of instructions per program
#define MAX_INS_NR                1000

// max. nr of instructions executed
#define MAX_INS_EXECUTED       1000000

// max. len of a label
#define MAX_LABEL_LEN               64

// max. len of a line in file with input datas
#define MAX_INPUT_LINE_LEN          64

//what to return when label[index] points 
//out of the code/to the non-data instruction
#define BAD_ARRAY_INDEX_VALUE        0

//and some error strings
#define ERR_USAGE "Usage:%s [-g] inputs!\n",argv[0]

typedef unsigned char byte;
typedef unsigned long ulng;

//LSB bit is used for num/reg,
//if the second LSB is set of, we must index, from
//  offset, either with num/reg
//num - value in "value"
//reg - reg nr. (0-25) in "value"
#define VALUE_TYPE_NUM   0
#define VALUE_TYPE_REG   1
#define VALUE_TYPE_INDEX 2

struct value {
  byte type;
  ulng value;
  ulng offset;
};

#define IF_TYPE_NONE    0
#define IF_TYPE_EQ      1
#define IF_TYPE_NE      2
#define IF_TYPE_LT      3
#define IF_TYPE_GT      4
#define IF_TYPE_LE      5
#define IF_TYPE_GE      6
#define IF_TYPE_NUM     7
#define IF_TYPE_DESC    {"","=","<>","<",">","<=",">="}
extern char if_type_desc[IF_TYPE_NUM][3];

//data have its value in "first.value"
//jmp in "first.offset"
#define INS_TYPE_ASSIGN 0
#define INS_TYPE_JMP    1
#define INS_TYPE_DATA   2
#define INS_TYPE_HALT   3

#define INS_OP_ASSIGN   0
#define INS_OP_ADD      1
#define INS_OP_SUB      2
#define INS_OP_MUL      3
#define INS_OP_DIV      4
#define INS_OP_MOD      5
#define INS_OP_AND      6
#define INS_OP_OR       7
#define INS_OP_XOR      8
#define INS_OP_SHL      9
#define INS_OP_SHR      10
#define INS_OP_ROL      11
#define INS_OP_ROR      12
#define INS_OP_NUM      13
#define INS_OP_DESC     {"","+","-","*","/","%","&","|","^","<<",">>","<<<",">>>"}
extern char ins_op_desc[INS_OP_NUM][4];

struct ins {
  byte if_type,type,op,reg;                  //op & reg - for ASSIGNMENT
  struct value if_first,if_second;
  struct value first, second;
};

extern struct ins prog[MAX_INS_NR];
extern int prog_len;
extern int max_ins_executed;


#endif
