#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "execute.h"

void debug_value(struct value *v) {
  switch(v->type) {
  case VALUE_TYPE_NUM:printf("%lu(#%lX) ",v->value,v->value);break;
  case VALUE_TYPE_REG:printf("%c ",(char) ('a'+v->value));break;
  case VALUE_TYPE_INDEX|VALUE_TYPE_NUM:printf("%lu[%lu] ",v->offset,v->value);break;
  case VALUE_TYPE_INDEX|VALUE_TYPE_REG:printf("%lu[%c] ",v->offset,(char) ('a'+v->value));break;
  }
}

void debug_line(ulng ins_executed,ulng ip) {
  printf("%8lu %8lu       ",ins_executed,ip);

  if (prog[ip].if_type) {
    printf("if ");
    debug_value(&prog[ip].if_first);
    printf("%s ",if_type_desc[prog[ip].if_type]);
    debug_value(&prog[ip].if_second);
  }

  switch(prog[ip].type) {
  case INS_TYPE_HALT:printf("halt ");break;
  case INS_TYPE_JMP:printf("jump %lu ",prog[ip].first.offset);break;
  case INS_TYPE_DATA:printf("data ");debug_value(&prog[ip].first);break;
  case INS_TYPE_ASSIGN:
    printf("%c = ",prog[ip].reg+'a');
    debug_value(&prog[ip].first);
    if (prog[ip].op!=INS_OP_ASSIGN) printf("%s ",ins_op_desc[prog[ip].op]);
    if (prog[ip].op!=INS_OP_ASSIGN) debug_value(&prog[ip].second);
  }
}

ulng regs[26];

//evaluated actual value of v
ulng get_value(struct value *v) {
  unsigned long long val;
  switch(v->type) {
  case VALUE_TYPE_NUM:return v->value;
  case VALUE_TYPE_REG:return regs[v->value];
  default:
    val=((v->type & 1) == VALUE_TYPE_NUM) ? v->value : regs[v->value];
    val+=v->offset;
    if (val < 0 || val >= prog_len || prog[val].type!=INS_TYPE_DATA) return BAD_ARRAY_INDEX_VALUE;
    return prog[val].first.value;
  }  
}

int execute(ulng input,int cmp_result,ulng result,int debug) {
  ulng ip=0;
  ulng ins_executed=0;

  //zeroed registers
  memset(regs,0,sizeof(regs));
  //input
  regs[0]=input;

  
  //while to run
  while (ip < prog_len && ins_executed < max_ins_executed) {
    ulng first,second;
    ulng rot=0;
    int if_passed=1;

    ins_executed++;
    if (debug) debug_line(ins_executed,ip);

    //any if?
    if (prog[ip].if_type) {
      first=get_value(&prog[ip].if_first);
      second=get_value(&prog[ip].if_second);
      switch(prog[ip].if_type) {
      case IF_TYPE_EQ:if (first!=second) if_passed=0;break;
      case IF_TYPE_NE:if (first==second) if_passed=0;break;
      case IF_TYPE_LT:if (first>=second) if_passed=0;break;
      case IF_TYPE_GT:if (first<=second) if_passed=0;break;
      case IF_TYPE_LE:if (first> second) if_passed=0;break;
      case IF_TYPE_GE:if (first< second) if_passed=0;break;
      }
      if (debug) printf("\t if %spassed ",(!if_passed) ? "not " : "");
    }

    if (if_passed)
    switch(prog[ip].type) {
    case INS_TYPE_HALT:ip=prog_len-1;break;
    case INS_TYPE_DATA:/*nop instruction*/break;
    case INS_TYPE_JMP:ip=prog[ip].first.offset-1;break;
    case INS_TYPE_ASSIGN:
      first=get_value(&prog[ip].first);
      second=get_value(&prog[ip].second);
      switch(prog[ip].op) {
      case INS_OP_ASSIGN:regs[prog[ip].reg]=first;break;
      case INS_OP_ADD:regs[prog[ip].reg]=first+second;break;
      case INS_OP_SUB:regs[prog[ip].reg]=first-second;break;
      case INS_OP_MUL:regs[prog[ip].reg]=first*second;break;
      case INS_OP_DIV:regs[prog[ip].reg]=(second) ? first/second : 0;break;
      case INS_OP_MOD:regs[prog[ip].reg]=(second) ? first%second : first;break;
      case INS_OP_AND:regs[prog[ip].reg]=first&second;break;
      case INS_OP_OR :regs[prog[ip].reg]=first|second;break;
      case INS_OP_XOR:regs[prog[ip].reg]=first^second;break;
      case INS_OP_SHL:regs[prog[ip].reg]=(second<32) ? first<<second : 0;break;
      case INS_OP_SHR:regs[prog[ip].reg]=(second<32) ? first>>second : 0;break;
      case INS_OP_ROL:rot=(32-(second&31))&31;goto ROTR_work;
      case INS_OP_ROR:rot=(second&31);
      ROTR_work:	
	if (rot) regs[prog[ip].reg]=(first>>rot) | (first<<(32-rot));
	else regs[prog[ip].reg]=first;
	break;
      }
      if (debug) printf("\t %c <- %lu(#%lX)",prog[ip].reg+'a',
			regs[prog[ip].reg],regs[prog[ip].reg]);
    }

    ip++;
    if (debug) printf("\n");
  }

  if (ip < prog_len && ins_executed >= max_ins_executed) {
    fprintf(stderr,"Program didn't terminate during the first %d instructions!\n",max_ins_executed);
    return 0;
  }

  if (cmp_result && regs[25]!=result) {
    fprintf(stderr,"The correct output for %lu is %lu and not %lu!\n",input,result,regs[25]);
    return 0;
  }

  if (!cmp_result)
    printf("For input %lu(#%lX) the program produced output %lu(#%lX).\n",
	   input,input,regs[25],regs[25]);
 
  return 1;
}
